open VecOps

type idx =
  [ `Char of (char * Sdlvideo.color)
  | `Block of (int * int * Sdlvideo.color)
  | `Bitmap of (string)
  | `Colored of (idx * Sdlvideo.color)
  | `Compose of (idx list)
  | `Move of ((int * int) * idx)
  | `Realpha of (float * idx)]

type t = { surfaces : (idx, ((int * int) * Sdlvideo.surface)) Hashtbl.t }

let font = lazy (Sdlttf.open_font "font.ttf" 25)

let get_face (ch, color) =
  let font_height = Sdlttf.font_height (Lazy.force font) in
  let s = Sdlttf.render_glyph_blended (Lazy.force font) ch color in
  let (minx, _maxx, _miny, maxy) = Sdlttf.glyph_metrics (Lazy.force font) ch in
    ((minx, font_height - maxy), s)

let create () =
  { surfaces = Hashtbl.create 100; }

let bitmaps = lazy (create ())

let get_bitmaps () = Lazy.force bitmaps

let collision_mask_of_surface s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Sdlvideo.get_pixel s x y <> 0l)
	    done )
      done );
    mask

let collision_mask_of_surface2 s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Int32.logand (Sdlvideo.get_pixel s x y) 0x00ffffffl = 0l)
	    done )
      done );
    mask


let map_surface f src =
  let w, h = Gfx.surface_dims src in
  let dst = Sdlvideo.create_RGB_surface_format src [] ~w ~h in
    GobaUtils.iter2d 
      (fun (x, y) ->
	 let ((r, g, b), alpha) = Sdlvideo.get_RGBA src (Sdlvideo.get_pixel src x y) in
	   Sdlvideo.put_pixel dst x y 
	     (let (r, g, b, alpha) = 
		f (r, g, b, alpha) in
		Sdlvideo.map_RGB dst ~alpha (r, g, b)
	     )
      )
      ((0, 0), (w, h));
    dst

(* recolors white/gray *)
let recolor (dr, dg, db) ((r, g, b, a) as orig) =
  let min = List.fold_left min r [g; b] in
  let max = List.fold_left max r [g; b] in
  let delta = max - min in
    if float delta /. float max < 0.05 then
      let br = float (List.fold_left (+) r [g; b]) /. (3.0 *. 255.0) in
      let color = (int_of_float (float dr *. br), 
		   int_of_float (float dg *. br),
		   int_of_float (float db *. br),
		   a)
      in
	color
    else
      orig

let realpha adjust (r, g, b, a) =
  (r, g, b, min 255 (max 0 (adjust a)))

let rec bitmap (idx : idx) =
  let bm = Lazy.force bitmaps in
    try Hashtbl.find bm.surfaces idx
    with Not_found ->
      let s = 
	match idx with
	  | `Char el -> get_face el
	  | `Block (width, height, color) ->
	      (0, 0), 
	      (let s = Sdlvideo.create_RGB_surface [] ~w:width ~h:height ~bpp:24 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0x0l in
		 Sdlvideo.fill_rect s (Sdlvideo.map_RGB s color);
		 s)
	  | `Bitmap (file) ->
	      (0, 0),
	      (Sdlloader.load_image (Printf.sprintf "gfx/%s.png" file))
	  | `Colored (idx, color) ->
	      let bm = bitmap idx in
		(fst bm, map_surface (recolor color) (snd bm))
	  | `Compose (idxs) -> compose idxs
	  | `Move (ofs, idx) ->
	      let bm = bitmap idx in
		(fst bm +| ofs, snd bm)
	  | `Realpha (adjust, idx) ->
	      let bm = bitmap idx in
		(fst bm, map_surface (realpha (fun a -> int_of_float (float a *. adjust))) (snd bm))
      in
	Hashtbl.add bm.surfaces idx s;
	s

and compose idxs =
  let bms = List.map bitmap idxs in
  let extreme f cmp = List.fold_left (fun v b -> if cmp v (f b) then v else (f b)) (f (List.hd bms)) (List.tl bms) in
  let min_x = extreme (fun ((x0, _), _) -> x0) (<) in
  let min_y = extreme (fun ((_, y0), _) -> y0) (<) in
  let max_x = extreme (fun ((x0, _), s) -> x0 + fst (Gfx.surface_dims s)) (>) in
  let max_y = extreme (fun ((_, y0), s) -> y0 + snd (Gfx.surface_dims s)) (>) in
  let ofs = min_x, min_y in
  let size = (max_x, max_y) -| (min_x, min_y) in
  let surface = Sdlvideo.create_RGB_surface [] ~w:(fst size) ~h:(snd size) ~bpp:24 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0x0l in
    List.iter 
      (fun (ofs', bm) ->
	 let ofs = ofs' -| ofs in
	 let size = Gfx.surface_dims bm in
	   Sdlvideo.blit_surface 
	     ~src:bm
	     ~dst:surface
	     ~dst_rect:(Sdlvideo.rect (fst ofs) (snd ofs) (fst size) (snd size))
	     ()
      )
      bms;
    ofs, surface

let dump_stats () =
  Printf.printf "Bitmaps stats: %d surfaces\n" (Hashtbl.length (Lazy.force bitmaps).surfaces)

let _ = at_exit dump_stats
