open VecOps
open GobaUtils

type t = { width	: int;
	   height	: int;
	   contents	: Tile.t array;
	   tile_size	: int * int;
	 }

let at map (x, y) =
  fst map.contents.(x + map.width * y)

let at_full map (x, y) =
  map.contents.(x + map.width * y)

let set map (x, y) value =
  map.contents.(x + map.width * y) <- value

let damage map at damage_hp =
  let ch, info = at_full map at in
  let real_damage =
    match ch with
      | Tile.Grass 
      | Tile.Mountain -> 
	  damage_hp
      | _ -> 0
  in
  let { Tile.hp = hp } = info in
  let new_damage = max 0 (info.Tile.hp - real_damage) in
  let delta = hp - new_damage in
    set map at (ch, { Tile.hp = new_damage });
    delta
    
let black = (0, 0, 0)

let yellow = (255, 255, 200)

let white = (255, 255, 255)

let grey = (128, 128, 128)

let dark_grey = (64, 64, 64)

let green = (0, 255, 0)

let blue = (128, 128, 255)

let brown = (139, 69, 19)

let clamp low high v = min high (max low v)

let scale low high v =
  float (v - low) /. float (high - low)

let ipolate3 v col1 col2 =
  let v = clamp 0.0 1.0 v in
  let s = Vector.v3 v in
  let s'rev = Vector.v3 (1.0 -. v) in
    Vector.int3 (s'rev *||. Vector.float3 col1 +||. s *||. Vector.float3 col2)

let color_map (tile, info) = 
  let hp = info.Tile.hp in
    match tile with
      | Tile.Space -> white
      | Tile.Grass -> ipolate3 (scale 10 0 hp) green black
      | Tile.Water -> blue
      | Tile.Mountain -> ipolate3 (scale 50 0 hp) grey dark_grey
      | Tile.Wall -> white
      | Tile.Start _ -> failwith "No color for start place"

let tile_at map map_at = 
  let mod_tile base =
    let xmod, ymod = fst map_at mod 3, snd map_at mod 3 in
      `Move ((0, 10), `Bitmap (Printf.sprintf "%s-%d%d" base xmod ymod))
  in
  let grass_tile () =
    mod_tile "grass"
  in
  let space_tile () = 
    mod_tile "space"
  in
    if fst map_at < 0 || snd map_at < 0 || fst map_at >= map.width || snd map_at >= map.height then
      `Char ('#', white)
    else
      let tile, info = at_full map map_at in
      let color = color_map (tile, info) in
	match tile with 
	  | Tile.Space -> space_tile ()
	  | Tile.Grass -> `Compose [space_tile (); `Realpha (float info.Tile.hp /. 10.0, grass_tile ())]
	  | _ -> `Compose [space_tile (); `Char (Tile.char_of_tile tile, color)]

let create 
    ?info
    { MapData.size = width, height; 
      tile_size = tile_size; 
      data = contents }  =
  { width	= width;
    height	= height;
    contents  = 
      (match info with
	 | None -> Array.map Tile.tile_with_info contents
	 | Some aux -> Array.mapi (fun i v -> (v, aux.(i))) contents);
    tile_size = tile_size;
  }

let block_cmask =
  lazy (let bitmap = Bitmaps.bitmap (`Block (20, 20, (255, 255, 255))) in
	Bitmaps.collision_mask_of_surface (snd bitmap))

class map_object map map_at tile_size at (collidable : bool) (destructible : bool) (friction : float) = 
object
  method get_state : float -> Sexplib.Sexp.t = failwith "Cannot get_state map_object"
  method set_state : Sexplib.Sexp.t -> float -> unit = failwith "Cannot set_state map_object"
  method self_owned = true
  method get_id = Object.MapId map_at
  method collision_area = 
    let ofs = (0.0, snd tile_size /. 2.0) in
      (at +|. ofs, at +|. ofs +|. tile_size), Lazy.force block_cmask
  method location = at
  method render (_gfx : Gfx.t) (_:Object.render_flags) (_at : int * int) = ()
  method impact (world : Object.World.t) (damage_hp : int) : unit = 
    let got_damage = damage map map_at damage_hp in
      world.Object.World.world_report_damage (Object.MapId map_at, got_damage);
    
  method collide (world : Object.World.t) = collidable

  method speed = 0.0, 0.0
  method set_speed (speed : float * float) = ()
  method collidable = collidable
  method destructible = destructible

  method friction = friction

  method cause_damage (_:int) = ()

  method is_alive = true

  method change = ()

  method tick (_:float) (_:Object.World.t) = ()

    (* temporary hack *)
  method set_self_owned (_:bool) = ()

  method set_location (at' : float * float) : unit = failwith "attempted to set a map tile coordinate"

  method changed_since (_:Object.serial) = false

  method test (_:Object.World.t) = ()

  method recharge = false
end

let real_coords map at =
  Vector.float2 (at *| map.tile_size)

let map_coords map at =
  at /|. Vector.float2 map.tile_size

let objects_near map ((x1, y1), (x2, y2)) : Object.t list =
  let objs = ref [] in
  let x1, y1 = map_coords map (x1, y1) in
  let x2, y2 = map_coords map (x2, y2) in
    for yc = (int_of_float (y1 -. 1.5)) to (int_of_float (y2 +. 0.0)) do
      for xc = (int_of_float (x1 -. 1.5)) to (int_of_float (x2 +. 0.0)) do
	let ch, info = 
	  if xc >= 0 && yc >= 0 && xc < map.width && yc < map.height then
	    at_full map(xc, yc)
	  else
	    Tile.tile_with_info (Tile.Wall)
	in
	let friction =
	  let hp = info.Tile.hp in
	  match ch with
	    | Tile.Grass when hp >= 5 -> 0.3
	    | Tile.Grass when hp >= 1 -> 0.15
	    | Tile.Grass -> 0.0
	    | Tile.Mountain -> 0.4
	    | _ -> 0.0
	in
	let collidable, destructible =
	  let hp = info.Tile.hp in
	  match ch with
	    | Tile.Mountain when hp > 50 -> true, true
	    | Tile.Mountain -> false, true
	    | Tile.Wall -> true, true
	    | Tile.Grass -> false, true
	    | _ -> false, false
	in
	  if info.Tile.hp > 0 && ch <> Tile.Space then
	    let obj = new map_object map (xc, yc) (Vector.float2 map.tile_size) (real_coords map (xc, yc)) collidable destructible friction in
	      objs := (obj :> Object.t) :: !objs
      done
    done;
    !objs

let render_map map (gfx:#Gfx.t) (flags:Object.render_flags) display_at (offset, size) (objects:#Object.t list) =
  gfx#with_clip_rect 
    (Sdlvideo.rect 
       (fst display_at) (snd display_at) 
       (int_of_float (fst size)) (int_of_float (snd size)))
    (fun () ->
       let map_offset = map_coords map offset in
       let map_size = Vector.int2 (size /|. Vector.float2 map.tile_size) in
       let offset_frac = 
	 let frac v = fst (modf v) in
	   frac (fst map_offset), frac (snd map_offset)
       in
       ( for x = -2 to fst map_size + 1 do
	     ( for y = -2 to snd map_size + 1 do
		   let coord = (x, y) in
		   let at = 
		     Vector.int2 
		       ((Vector.float2 coord -|. offset_frac) *|. Vector.float2 map.tile_size) in
		   let map_at = coord +| Vector.int2 map_offset in
		   let (surface_offset, surface) = Bitmaps.bitmap (tile_at map map_at) in
		     gfx#blit (at +| surface_offset +| display_at) surface
	       done)
	 done);

	 (* Draw objects using painter's algorithm *)
	 let objs =
	   List.sort
	     (fun obj1 obj2 ->
		compare (snd (obj1#location)) (snd (obj2#location)))
	     objects
	 in
	   List.iter (fun obj -> 
			obj#render (gfx :> Gfx.t) flags
			  (int_of_float2 
			     ((obj#location -|. offset) +|. float2 (display_at))
			  )
		     ) objs
    )
