open VecOps

let bitmap = lazy (Bitmaps.bitmap (`Block (5, 5, (255, 0, 0))))
let cmask = lazy (Bitmaps.collision_mask_of_surface (snd (Lazy.force bitmap)))

let hit_sound = lazy (Sdlmixer.loadWAV "hit.wav")

let orientation_speed = 10.0

let fire_sound = lazy (Sdlmixer.loadWAV "ampu.wav")

class t id init constr =
  let orientation = constr.Constr.proj_orientation in
  let parent = constr.Constr.proj_parent in
  let normal = Vector.coaxel (Vector.unit orientation) in
  let default = Object.default init ((0.0, 0.0), (0.0, 0.0)) in

object (self)
  inherit Object.t id

  val mutable at = fst default
  val mutable speed = snd default
  val mutable alive = true

  val mutable distance = 90.0

  method location = at

  method speed = speed
  method set_speed speed' = speed <- speed'

  method set_location at' = at <- at'

  method is_alive = alive

  method render (gfx:#Gfx.t) _ at =
    if alive then
      let ofs, surface = Lazy.force bitmap in
	gfx#blit (at +| ofs) surface

  method detonate explosion =
    begin
      match explosion with
	| Some (world, direction) -> 
	    world.Object.World.world_explode at direction;
	    Sdlmixer.play_channel ~channel:6 (Lazy.force hit_sound);
	| None -> ();
    end;
    alive <- false;
    self#change;

  method tick delta world =
    if alive then
      begin
	distance <- 
	  distance -. 
	      Vector.abs2 (speed /|. (1.0, (* tank viewport ratio *) 180.0 /. 195.0)) *. delta;
	if distance < 0.0 then
	  self#detonate (Some (world, (0.0, 0.0)))
	else
	  begin
	    self#change;
	    at <- at +|. speed *|. (delta, delta);
	    speed <- speed -|. 
		normal *|. Vector.dup (Vector.dot speed normal *. delta *. orientation_speed);
	    match world.Object.World.world_collision (self :> Collision.t) (fun o -> o#get_id <> parent && o#destructible) with
	      | None -> ()
	      | Some (x, objs) -> 
		  if self#self_owned then
		    let rec give_damages damage_left damages =
		      let damages, _, damage_left =
			List.fold_left
			  (fun (left, right, damage_left) obj ->
			     let r = List.hd right in
			       if damage_left > 0 then
				 (r + 1::left, List.tl right, damage_left - 1)
			       else
				 (r::left, List.tl right, damage_left)
			  )
			  ([], damages, damage_left)
			  objs
		      in
		      let damages = List.rev damages in
			if damage_left > 0 then
			  give_damages damage_left damages
			else
			  damages
		    in
		    let damages = give_damages 5 (List.map (fun _ -> 0) objs) in
		    let l = Vector.abs2 x in
		    let u = (Vector.unit x -|. Vector.unit speed) /|. Vector.dup 2.0 in
		      List.iter2 (fun obj damage -> obj#impact world damage) objs damages; 
		      self#detonate (Some (world, u *|. Vector.dup l))
		  else
		    self#detonate (Some (world, (0.0, 0.0)))
	  end
      end

  method collision_area =
    (at, at +|. (5.0, 5.0)), Lazy.force cmask

  method cause_damage _ =
    self#detonate None

  method collidable = false

  initializer
    (match init with
       | Object.Init _ -> ()
       | Object.State (state0, age) -> self#set_state state0 age);
    self#change;
    Sdlmixer.play_channel ~channel:4 (Lazy.force fire_sound);
end
